-------------------------------------------------------------------------------
-- Localized Lua globals
-------------------------------------------------------------------------------
local _G = getfenv(0)

local string = _G.string
local math = _G.math
local table = _G.table

local tostring = _G.tostring
local time = _G.time
local type = _G.type

local wipe = table.wipe
local tinsert = table.insert
local tremove = table.remove
local tsort = table.sort

local ipairs, pairs = _G.ipairs, _G.pairs

-------------------------------------------------------------------------------
-- Localized Blizzard API
-------------------------------------------------------------------------------
local UnitClass, UnitLevel, UnitName = _G.UnitClass, _G.UnitLevel, _G.UnitName
local GetItemInfo, GetItemGem = _G.GetItemInfo, _G.GetItemGem
local GetInventoryItemLink = _G.GetInventoryItemLink

-------------------------------------------------------------------------------
-- AddOn namespace
-------------------------------------------------------------------------------
local MODNAME = "GearGauge"
local GearGauge = LibStub("AceAddon-3.0"):NewAddon(MODNAME, "AceEvent-3.0", "AceConsole-3.0")

_G[MODNAME] = GearGauge

local LDB = LibStub("LibDataBroker-1.1")
local QTC = LibStub("LibQTipClick-1.1")
local LBC = LibStub("LibBabble-Class-3.0"):GetLookupTable()
local LIB = LibStub("LibItemBonus-2.0")

local dev = false
--[===[@debug@
dev = true
--@end-debug@]===]
local L = LibStub("AceLocale-3.0"):GetLocale(MODNAME, "enUS", true, dev)

local DataObj
local tooltip

local version = GetAddOnMetadata(MODNAME, "Version")
version = string.gsub(version, "@project.version@", " - Development")

-------------------------------------------------------------------------------
-- Constants
-------------------------------------------------------------------------------
local TBC_MAXLEV = 60
local WOTLK_MAXLEV = 71
local IS_ENCHANTER = false

local COLOR_RED = "|cffff0000"

local EQUIPMENT = {
	["Head"]		= { minlev = 30,	name = _G["INVTYPE_HEAD"] },
	["Neck"]		= { minlev = 30,	name = _G["INVTYPE_NECK"] },
	["Shoulder"]		= { minlev = 20,	name = _G["INVTYPE_SHOULDER"] },
	["Shirt"]		= { minlev = 1,		name = _G["INVTYPE_BODY"] },
	["Chest"]		= { minlev = 1,		name = _G["INVTYPE_CHEST"] },
	["Waist"]		= { minlev = 1,		name = _G["INVTYPE_WAIST"] },
	["Legs"]		= { minlev = 1,		name = _G["INVTYPE_LEGS"] },
	["Feet"]		= { minlev = 1,		name = _G["INVTYPE_FEET"] },
	["Wrist"]		= { minlev = 1,		name = _G["INVTYPE_WRIST"] },
	["Hands"]		= { minlev = 1,		name = _G["INVTYPE_HAND"] },
	["Finger0"]		= { minlev = 20,	name = _G["INVTYPE_FINGER"] },
	["Finger1"]		= { minlev = 20,	name = _G["INVTYPE_FINGER"] },
	["Trinket0"]		= { minlev = 50,	name = _G["INVTYPE_TRINKET"] },
	["Trinket1"]		= { minlev = 50,	name = _G["INVTYPE_TRINKET"] },
	["Back"]		= { minlev = 1,		name = _G["INVTYPE_CLOAK"] },
	["MainHand"]		= { minlev = 1,		name = _G["INVTYPE_WEAPONMAINHAND"] },
	["SecondaryHand"]	= { minlev = 1,		name = _G["INVTYPE_WEAPONOFFHAND"] },
	["Ranged"]		= { minlev = 1,		name = _G["INVTYPE_RANGED"] },
	["Tabard"]		= { minlev = 1,		name = _G["INVTYPE_TABARD"] },
}

local ORDERED_EQUIPMENT = {
	[1]	= "Head",	[2]	= "Neck",		[3]	= "Shoulder",
	[4]	= "Shirt",	[5]	= "Chest",		[6]	= "Waist",
	[7]	= "Legs",	[8]	= "Feet",		[9]	= "Wrist",
	[10]	= "Hands",	[11]	= "Finger0",		[12]	= "Finger1",
	[13]	= "Trinket0",	[14]	= "Trinket1",		[15]	= "Back",
	[16]	= "MainHand",	[17]	= "SecondaryHand",	[18]	= "Ranged",
	[19]	= "Tabard"
}

local TWO_HANDED_WEAPON = {
	[L["Fishing Poles"]] = true,
	[L["Polearms"]] = true,
	[L["Staves"]] = true,
	[L["Two-Handed Axes"]] = true,
	[L["Two-Handed Maces"]] = true,
	[L["Two-Handed Swords"]] = true,
}

local GEM_TYPES = {
	"Blue",
	"Meta",
	"Red",
	"Yellow"
}

local BONUS_VALUE = {
	-------------------------------------------------------------------------------
	-- Enchantments
	-------------------------------------------------------------------------------
	["ARMOR_BONUS"]			= 0.01,		["DEFENSE"]			= 0.25,
	["PARRY"]			= 1,		["SPIRIT"]			= 0.2,

	-------------------------------------------------------------------------------
	-- Stats
	-------------------------------------------------------------------------------
	["AGI"]				= 0.2,		["BASE_ARMOR"]			= 0.01,
	["INT"]				= 0.2,		["SPI"]				= 0.2,
	["STA"]				= 0.2,		["STR"]				= 0.2,

	-------------------------------------------------------------------------------
	-- Power
	-------------------------------------------------------------------------------
	["ATTACKPOWER"]			= 1,		["ATTACKPOWERBEAST"]		= 0.1,
	["ATTACKPOWERDEMON"]		= 0.1,		["ATTACKPOWERDRAGON"]		= 0.1,
	["ATTACKPOWERELEMENTAL"]	= 0.1,		["ATTACKPOWERUNDEAD"]		= 0.1,
	["SPELLPOWER"]			= 1,

	-------------------------------------------------------------------------------
	-- Combat ratings.
	-------------------------------------------------------------------------------
	["CR_ARMOR_PENETRATION"]	= 0.1,		["CR_BLOCK"]			= 1,
	["CR_CRIT"]			= 1,		["CR_DEFENSE"]			= 0.25,
	["CR_DODGE"]			= 0.25,		["CR_EXPERTISE"]		= 0.1,
	["CR_HASTE"]			= 1,		["CR_HIT"]			= 1,
	["CR_PARRY"]			= 1,		["CR_RANGEDCRIT"]		= 1,
	["CR_RANGEDHASTE"]		= 1,		["CR_RANGEDHIT"]		= 1,
	["CR_RESILIENCE"]		= 0.1,

	-------------------------------------------------------------------------------
	-- Base values
	-------------------------------------------------------------------------------
	["BLOCKVALUE"]			= 1,		["HEALTH"]			= 0.1,
	["MANA"]			= 0.1,

	-------------------------------------------------------------------------------
	-- Regeneration
	-------------------------------------------------------------------------------
	["HEALTHREG"]			= 0.35,		["MANAREG"]			= 0.35,

	-------------------------------------------------------------------------------
	-- Resistances
	-------------------------------------------------------------------------------
	["ARCANERES"]			= 0.125,	["FIRERES"]			= 0.125,
	["FROSTRES"]			= 0.125,	["HOLYRES"]			= 0.125,
	["NATURERES"]			= 0.125,	["SHADOWRES"]			= 0.125,

	-------------------------------------------------------------------------------
	-- Percentage modifiers
	-------------------------------------------------------------------------------
	["MOD_ARMOR"]			= 1,		["MOD_DAMAGE_CRIT"]		= 1,
	["MOD_HEAL_CRIT"]		= 1,		["MOD_INT"]			= 1,
	["MOD_MANA"]			= 1,		["MOD_SPELLDAMAGE_TAKEN"]	= 1,
	["MOD_STA"]			= 1,

	-------------------------------------------------------------------------------
	-- Reductions
	-------------------------------------------------------------------------------
	["FEARREDUCTION"]		= 0.1,		["DISARMREDUCTION"]		= 0.1,
	["SILENCEREDUCTION"]		= 0.1,		["SNAREREDUCTION"]		= 0.1,
	["STUNREDUCTION"]		= 0.1,		["THREATREDUCTION"]		= 1,

	-------------------------------------------------------------------------------
	-- Miscellaneous
	-------------------------------------------------------------------------------
	["RUNSPEED"]			= 0.1,		["SPELLPEN"]			= 0.1,
}

local CLASS_NUMBER = {
	["DEATHKNIGHT"]	= 1,
	["DRUID"]	= 2,
	["HUNTER"]	= 3,
	["MAGE"]	= 4,
	["PALADIN"]	= 5,
	["PRIEST"]	= 6,
	["ROGUE"]	= 7,
	["SHAMAN"]	= 8,
	["WARLOCK"]	= 9,
	["WARRIOR"]	= 10,
}

local ORDERED_CLASSES = {
	[1]	= "DEATHKNIGHT",
	[2]	= "DRUID",
	[3]	= "HUNTER",
	[4]	= "MAGE",
	[5]	= "PALADIN",
	[6]	= "PRIEST",
	[7]	= "ROGUE",
	[8]	= "SHAMAN",
	[9]	= "WARLOCK",
	[10]	= "WARRIOR",
}


-------------------------------------------------------------------------------
-- Variables
-------------------------------------------------------------------------------
local player_equip		-- Player's equipment list.
local db

local function debug(msg)
	--[===[@debug@
	print(string.format("%sDebug|r: %s", COLOR_RED, tostring(msg)))
	--@end-debug@]===]
end

-------------------------------------------------------------------------------
-- Table cache
-------------------------------------------------------------------------------
local used_tables = 0
local cached_tables = 0
--[===[@debug@
local table_locs = {}
--@end-debug@]===]

local AcquireTable, ReleaseTable
do
	local table_cache = {}

	-- Returns a table
	function AcquireTable(where)
		local tbl = tremove(table_cache) or {}
		used_tables = used_tables + 1
		cached_tables = #table_cache
		--[===[@debug@
		table_locs[tbl] = where
		--@end-debug@]===]
		return tbl
	end

	-- Cleans the table and stores it in the cache.
	function ReleaseTable(tbl)
		if not tbl then return false end

		wipe(tbl)
		tinsert(table_cache, tbl)
		used_tables = used_tables - 1
		cached_tables = #table_cache
		--[===[@debug@
		table_locs[tbl] = nil
		--@end-debug@]===]
		return true
	end
end	-- do

-------------------------------------------------------------------------------
-- Local functions
-------------------------------------------------------------------------------
-- Finds an entry in the entries table by name.
local function FindEntry(entry)
	local found_name, search_name
	entry = string.lower(entry)

	for name in pairs(db.realm.entries) do
		search_name = string.lower(name)

		if search_name == entry then
			found_name = name
			break;
		end
	end
	return found_name
end

-- Releases all of the sub-tables in the equipment list, then releases the list.
local function ReleaseItems(item_list)
	if not item_list then
		debug("No item list.")
		return
	end

	local count = 0
	for idx, slot in ipairs(ORDERED_EQUIPMENT) do
		if ReleaseTable(item_list[slot]) then
			count = count + 1
		end
	end
	if ReleaseTable(item_list) then
		count = count + 1
	end
end

local function AverageItemLevels(equip_list)
	local level_cnt = 0
	local item_cnt = 0

	for slot in pairs(EQUIPMENT) do
		local item = equip_list[slot]

		if item.level > 0 and slot ~= "Shirt" and slot ~= "Tabard" then
			level_cnt = level_cnt + item.level
			item_cnt = item_cnt + 1
		end
	end
	if item_cnt == 0 then return 0 end

	local avg = math.floor(level_cnt / item_cnt)
	return avg
end

local function CanScan(unit, silent)
	if unit == "player" or UnitName(unit) == UnitName("player") then return false end

	local enemies = UnitFactionGroup(unit) ~= UnitFactionGroup("player")
	local prefs = db.global.scanning

	if enemies and not prefs.enemies then return false end

	if InCombatLockdown() or not CheckInteractDistance(unit, 1)
		or not (UnitExists(unit) and UnitIsPlayer(unit)
			and (UnitLevel(unit) >= prefs.minlevel) and (UnitLevel(unit) <= prefs.maxlevel)) then return false end

	if not CanInspect(unit, false) then
		if not silent then
			GearGauge:Print(string.format("%s%s: %s|r", COLOR_RED, UnitName(unit), ERR_PLAYER_WRONG_FACTION))
		end
		return false
	end
	return true
end

-------------------------------------------------------------------------------
-- Item status generation
-------------------------------------------------------------------------------
local BuildGearData
do
	-------------------------------------------------------------------------------
	-- Check to see if an empty slot is OK given the unit's level/class.
	-------------------------------------------------------------------------------
	local function IsEmptySlotOK(slot, level, class)
		if slot == "Shirt" or slot == "Tabard" then return true end

		if slot == "SecondaryHand"  then
			if level < 5 and (class == LBC["Paladin"] or class == LBC["Shaman"] or class == LBC["Warrior"]) then	-- Shields
				return true
			else
				return false
			end
		end

		if slot == "Ranged" then
			if  level < 52 and (class == LBC["Druid"] or class == LBC["Paladin"] or class == LBC["Shaman"]) then	-- Idols, Librams, Totems
				return true
			elseif level < 5 and (class == LBC["Mage"] or class == LBC["Priest"] or class == LBC["Warlock"]) then	-- Wands
				return true
			else
				return false
			end
		end
		return EQUIPMENT[slot].minlev > level
	end

	-- Item subtypes which are not enchantable
	local NON_ENCHANTABLE = {
		[L["Idols"]] = true,
		[L["Librams"]] = true,
		[L["Miscellaneous"]] = true,
		[L["Other"]] = true,
		[L["Sigils"]] = true,
		[L["Thrown"]] = true,
		[L["Totems"]] = true,
		[L["Wands"]] = true,
	}
	local prev_stype	-- Placeholder in case previous slot contained a two-handed weapon.
	local current_type	-- Type of "unit" passed to BuildGearData() and ConstructItem().

	-------------------------------------------------------------------------------
	-- Assign values to an item for later use.
	-------------------------------------------------------------------------------
	local function ConstructItem(item_link, slot, unit)
		local unit_level, unit_class
		local item = AcquireTable("ConstructItem")

		if current_type ~= "table" then
			unit_level = UnitLevel(unit)
			unit_class = UnitClass(unit)
		else
			unit_level = unit.level
			unit_class = unit.class
		end
		-- Default values here
		item.slot, item.rating, item.level = slot, 0, 0
		item.emptyMeta, item.emptyRed, item.emptyYellow, item.emptyBlue = 0, 0, 0, 0
		item.ench_id = -1

		if not item_link then
			local two_handed = slot == "SecondaryHand" and TWO_HANDED_WEAPON[prev_stype]
			local is_ok = IsEmptySlotOK(slot, unit_level, unit_class)

			if not two_handed and not is_ok then
				item.rating = 0 - (unit_level * 0.5)
			end
			item.empty_ok = is_ok
			prev_stype = nil
			return item
		end
		item.name, item.link, item.rarity, item.level, item.minlev, item.type, item.stype, item.stacksize, item.equiploc, item.texture = GetItemInfo(item_link)
		prev_stype = item.stype

		-- Set the level of Heirloom items depending on the player's level
		if item.rarity == 7 then
			if unit_level <= TBC_MAXLEV then
				item.level = unit_level
			elseif unit_level > TBC_MAXLEV and unit_level <= WOTLK_MAXLEV then
				item.level = 3 * (unit_level - 64) + 90
			elseif unit_level > WOTLK_MAXLEV then
				item.level = 4 * (unit_level - 73) + 142
			end
		end
		if not item.level then item.level = 0 end	-- Shouldn't happen, but it did.
		-- Must check for Tabard and Shirt here since they may have green text which LibItemBonus cannot properly parse, yet will still allocate a bonus table for.
		if item.level <= 1 or slot == "Tabard" or slot == "Shirt" then return item end

		item.info = LIB:ScanItemLink(item_link)

		local points = 0

                if item.info.bonuses then
			-- Get gem information
			item.emptyBlue = item.info.bonuses["EMPTY_SOCKET_BLUE"] or 0
			item.info.bonuses["EMPTY_SOCKET_BLUE"] = nil

			item.emptyMeta = item.info.bonuses["EMPTY_SOCKET_META"] or 0
			item.info.bonuses["EMPTY_SOCKET_META"] = nil

			item.emptyRed = item.info.bonuses["EMPTY_SOCKET_RED"] or 0
			item.info.bonuses["EMPTY_SOCKET_RED"] = nil

			item.emptyYellow = item.info.bonuses["EMPTY_SOCKET_YELLOW"] or 0
			item.info.bonuses["EMPTY_SOCKET_YELLOW"] = nil

			-- Remove these - they are duplicated.
			item.info.bonuses["CR_SPELLHASTE"] = nil
			item.info.bonuses["CR_SPELLCRIT"] = nil
			item.info.bonuses["CR_SPELLHIT"] = nil

			-- Remove these
			item.info.bonuses["WEAPON_SPEED"] = nil
			item.info.bonuses["WEAPON_MIN"] = nil
			item.info.bonuses["WEAPON_MAX"] = nil

                        for k, v in pairs(item.info.bonuses) do
				local bonus = LIB:GetBonus(k, item.info.bonuses, unit_level)

				if BONUS_VALUE[k] then bonus = bonus * BONUS_VALUE[k] end

				points = points + bonus
                        end
                end
		points = ((points + (item.level / unit_level)) * 0.1)

		local rating = points
		local _, _, ench_id = string.split(":", item_link)		-- Grab the item's enchant id, if enchanted.

		if slot ~= "Waist" and slot ~= "Shirt" and slot ~= "Tabard" then
			item.ench_id = ench_id

			if ench_id == nil or ench_id == "0" then
				if ((slot == "Finger0" or slot == "Finger1") and (unit == "player" and IS_ENCHANTER)) or not NON_ENCHANTABLE[item.stype] then
					rating = rating - (points * 0.25)
				else
					item.ench_id = -1
				end
			end
		end

		for k, v in ipairs(GEM_TYPES) do
			if item["empty"..v] > 0 then
				rating = rating - ((points * 0.125) * item["empty"..v])
			end
		end
		item.rating = rating

		return item
	end

	-------------------------------------------------------------------------------
	-- List of items equipped by unit.
	-------------------------------------------------------------------------------
	function BuildGearData(unit, item_list, link_list)
		local rating = 0

		current_type = type(unit)	-- Set the current unit type for ConstructItem().
		prev_stype = nil	-- Nil this out - it's used in ConstructItem().

		for idx, slot in ipairs(ORDERED_EQUIPMENT) do
			local link

			if current_type ~= "table" then
				link = GetInventoryItemLink(unit, EQUIPMENT[slot].id)
			else
				link = unit.link_list[slot]
			end
			local item = ConstructItem(link, slot, unit)

			rating = rating + item.rating
			item_list[slot] = item

			if link_list then
				link_list[slot] = link
			end
		end
		return rating
	end
end -- do

do
	local function CreatePlayerInfo(unit)
		local info = AcquireTable("CreatePlayerInfo")
		local local_name, class_name = UnitClass(unit)

		db.global.localized_classes[class_name] = local_name

		info.class = CLASS_NUMBER[class_name]
		info.faction = UnitFactionGroup(unit)
		info.guild = GetGuildInfo(unit)
		info.link_list = AcquireTable("CreatePlayerInfo (link_list)")
		info.level = UnitLevel(unit)
		info.race = UnitRace(unit)
		return info
	end

	function GearGauge:Scan(unit)
		if not CanScan(unit, false) then return end
		NotifyInspect(unit)

		local prefs = db.global.scanning
		local unit_name = UnitName(unit)
		local target = db.realm.entries[unit_name]

		if not target then
			target = CreatePlayerInfo(unit)
			target.scanned = time()
			db.realm.entries[unit_name] = target
			self:RefreshList()
		else
			local scanned = time() - target.scanned

			-- If the delay hasn't been reached yet, don't rescan.
			if scanned < (60 * prefs.delay) then
				return
			else
				target.scanned = time()
			end
		end
		local item_list = AcquireTable("GearGauge:Scan")
		local rating = BuildGearData(unit, item_list, target.link_list)
		local equip_level = AverageItemLevels(item_list)

		ReleaseItems(item_list)

		if not prefs.notify then return end

		local color_str = ""
		if UnitFactionGroup(unit) ~= UnitFactionGroup("player") then color_str = COLOR_RED end

		self:Print(string.format("%s%s %s: %.2f. %s %s: %d", color_str, unit_name, RATING, rating, ITEMS, LEVEL, equip_level))
	end
end -- do

local function RemoveEntry(entry, notify)
	ReleaseTable(db.realm.entries[entry].link_list)
	ReleaseTable(db.realm.entries[entry])
	db.realm.entries[entry] = nil

	if notify then
		GearGauge:Print(string.format(L["ENTRY_DELETE_MSG"], entry))
	end
end

local function CheckExpiration()
	local scanned
	local persistence = db.global.scanning.persistence
	local count = 0

	for entry in pairs(db.realm.entries) do
		scanned = time() - db.realm.entries[entry].scanned

		if scanned >= (86400 * persistence) then
			count = count + 1
			RemoveEntry(entry)
		end
	end
	return count
end

-- Rebuild the player's gear rating, and update the DataObj text.
local function UpdateData()
	ReleaseTable(player_equip)
	player_equip = AcquireTable("UpdateData")

	DataObj:SetIcon()
	DataObj.text = string.format("%.2f", BuildGearData("player", player_equip, nil))
end

-------------------------------------------------------------------------------
-- Tooltip functions
-------------------------------------------------------------------------------
local tipdata = {}	-- So the tooltip can be refreshed elsewhere.

local ShowTooltip
do
	-- Returns a string based on whether or not the item is enchanted
	local function GetItemStatus(item)
		if not item.link then return HELPFRAME_ITEM_TITLE end

		local ench_id = item.ench_id
		local retstr = ""

		if ench_id ~= -1 and (ench_id == nil or ench_id == "0") then
			retstr = retstr.."|TInterface\\Icons\\Trade_Engraving:22:22|t"
		end
		local socket_tex = "|TInterface\\ItemSocketingFrame\\UI-EmptySocket-"

		for k, v in ipairs(GEM_TYPES) do
			if item["empty"..v] > 0 then
				retstr = retstr..socket_tex..v..":22:22|tx"..item["empty"..v]
			end
		end
		return retstr
	end
	local QTip = LibStub("LibQTip-1.0")
	local bonus_tip

	local function ItemOnEnter(event, cell, arg)
		if not arg then return end
		QTC.OnEnter(event, cell, arg)

		bonus_tip = QTip:Acquire(MODNAME.."BonusTip", 2, "LEFT", "LEFT")
		bonus_tip:ClearAllPoints()
		bonus_tip:SetPoint("TOPRIGHT", tooltip, "TOPLEFT")
		bonus_tip:Clear()

		local line, column = bonus_tip:AddLine()
		local count = 0
		bonus_tip:SetCell(line, column, L["Item Bonuses"], "CENTER", 2)
		bonus_tip:AddSeparator()

		for k, v in pairs(arg) do
			bonus_tip:AddLine(BONUS_VALUE[k] and L[k] or k, v)
			count = count + 1
		end

		if count == 0 then
			line, column = bonus_tip:AddLine()
			bonus_tip:SetCell(line, column, UNKNOWN, "CENTER", 2)
		end
		if TipTac and TipTac.AddModifiedTip then TipTac:AddModifiedTip(bonus_tip, true) end

		bonus_tip:Show()
	end

	local function ItemOnLeave(event, cell, arg)
		if not bonus_tip then return end
		QTC.OnLeave(event, cell, arg)
		bonus_tip = QTip:Release(bonus_tip)
	end

	local function NilMouse()
	end

	function ShowTooltip(anchor, unit, center, refresh)
		local prefs = db.global.tooltip

		tooltip = QTC:Acquire(MODNAME.."Tooltip", 3, "LEFT", "LEFT", "LEFT")
		tooltip:SetCallback("OnEnter", ItemOnEnter)
		tooltip:SetCallback("OnLeave", ItemOnLeave)
		tooltip:SetCallback("OnMouseDown", NilMouse)
		tooltip:SetAutoHideDelay(prefs.timer, anchor)

		if not refresh then
			tooltip:ClearAllPoints()

			if not center then
				tooltip:SmartAnchorTo(anchor)
			else
				tooltip:SetPoint("CENTER", anchor, 0, 0)
			end
		end
		tooltip:Clear()

		if TipTac and TipTac.AddModifiedTip then TipTac:AddModifiedTip(tooltip, true) end

		local item_list
		local unit_name, unit_class, unit_race, unit_level, unit_guild

		unit = unit or "player"

		-- This is set so that the tooltip can be refreshed elsewhere.
		tipdata.anchor = anchor
		tipdata.unit = unit
		tipdata.center = center

		if unit == "player" then
			item_list = player_equip
			unit_name = UnitName(unit)
			unit_class = UnitClass(unit)
			unit_race = UnitRace(unit)
			unit_level = UnitLevel(unit)
			unit_guild = GetGuildInfo(unit)
		else
			item_list = AcquireTable("ShowTooltip")

			if unit == "target" then
				unit_name = UnitName("target")
			else
				unit_name = unit	-- Not player or target, so name was passed as unit.
			end

			local entry = db.realm.entries[unit_name]
			BuildGearData(entry, item_list)
			unit_level = entry.level
			unit_class = entry.class
			unit_race = entry.race
			unit_guild = entry.guild
		end
		local rating_tot = 0
		local line, column

		line, column = tooltip:AddNormalHeader()
		tooltip:SetNormalCell(line, column, unit_name.." - "..LEVEL.." "..unit_level.." "..unit_race.." "..db.global.localized_classes[ORDERED_CLASSES[unit_class]], "CENTER", 3)

		if unit_guild then
			line, column = tooltip:AddNormalHeader()
			tooltip:SetNormalCell(line, column, string.format("<%s>", unit_guild), "CENTER", 3)
		end
		tooltip:AddSeparator()
		line, column = tooltip:AddNormalHeader("", RATING, ADDON_MISSING)

		tooltip:SetNormalCell(line, 1, ITEMS, "CENTER")
		tooltip:AddSeparator()

		local two_handed = nil
		local prev_stype = nil

		for idx, slot in ipairs(ORDERED_EQUIPMENT) do
			local item = item_list[slot]

			rating_tot = rating_tot + item.rating

			if item.link then
				local r, g, b, hex = GetItemQualityColor(item.rarity)

				line, column = tooltip:AddNormalLine()

				if prefs.coloring == "Line" or prefs.coloring == "Both" then
					tooltip:SetLineColor(line, r, g, b)
				end
				local item_name = item.name

				if prefs.coloring == "Text" or prefs.coloring == "Both" then
					item_name = hex..item_name.."|r"
				end

				if not item.info or not item.info.bonuses then
					tooltip:SetNormalCell(line, 1, "|T"..item.texture..":22:22|t"..item_name)
				else
					tooltip:SetCell(line, 1, "|T"..item.texture..":22:22|t"..item_name, item.info.bonuses)
				end
				tooltip:SetNormalCell(line, 2, string.format("%.2f", item.rating))
				tooltip:SetNormalCell(line, 3, GetItemStatus(item))
				prev_stype = item.stype
			else
				local two_handed = slot == "SecondaryHand" and TWO_HANDED_WEAPON[prev_stype]

				if not two_handed  then
					line, column = tooltip:AddLine()

					if not item.empty_ok then
						tooltip:SetLineColor(line, 1)
					end
					tooltip:SetNormalCell(line, 1, EQUIPMENT[slot].name, nil, nil, 1, nil, 17)
					tooltip:SetNormalCell(line, 2, string.format("%.2f", item.rating))
					tooltip:SetNormalCell(line, 3, GetItemStatus(item))
					prev_stype = nil
				end
			end
		end
		tooltip:AddSeparator()

		line, column = tooltip:AddLine()

		tooltip:SetNormalCell(line, 1, string.format("%s: %d", LEVEL, AverageItemLevels(item_list)))
		tooltip:SetNormalCell(line, 2, string.format("%.2f", rating_tot))
		tooltip:Show()

		if unit ~= "player" then
			ReleaseItems(item_list)
		end
	end
end	-- do

-------------------------------------------------------------------------------
-- Initialization functions
-------------------------------------------------------------------------------
local OpenConfig
local DB_VERSION = 0.1

function GearGauge:OnInitialize()
	-- Set the numeric ID and texture strings for the equipment slots
	for slot in pairs(EQUIPMENT) do
		EQUIPMENT[slot].id, EQUIPMENT[slot].texture = GetInventorySlotInfo(slot.."Slot")
	end

	-- Determine if the toon is an enchanter
	local enchanting_sklname = GetSpellInfo(7411)
	for idx = 1, GetNumSkillLines() do
		local skill = GetSkillLineInfo(idx)
		if skill == enchanting_sklname then
			IS_ENCHANTER = true
			break
		end
	end
	DataObj = LDB:NewDataObject(MODNAME, {
		type	= "data source",
		label	= MODNAME,
		text	= "0",
		SetIcon = function(self)
				  local num = math.random(8)
				  self.icon = "Interface\\Icons\\INV_Misc_Gear_0"..num
			  end,
		OnEnter = function(display, motion) ShowTooltip(display) end,
		OnLeave = function(display, motion) end,
		OnClick = function(button) OpenConfig() end,
	})
	DataObj:SetIcon()

	local defaults = {
		global = {
			minimap_icon		= {
				hide		= true,
			},
			scanning = {
				delay		= 10,
				persistence	= 30,
				maxlevel	= MAX_PLAYER_LEVEL,
				minlevel	= 1,
				notify		= true,
				enemies		= true,
			},
			tooltip = {
				coloring	= "Both",
				timer		= 0.25,
			},
			-- This is filled in as we go. Yes, it is hackish.
			localized_classes = {},
		},
		realm = {
			db_version = 0,
			entries = {},		-- Scanned players.
		},
	}
	db = LibStub("AceDB-3.0"):New(MODNAME.."DB", defaults)

	LibStub("LibDBIcon-1.0"):Register(MODNAME, DataObj, db.global.minimap_icon)
	self:SetupOptions()
	self:RegisterChatCommand("geargauge", "SlashCommand")
	self:RegisterChatCommand("gg", "SlashCommand")
	self.OnInitialize = nil
end

function GearGauge:CheckDatabaseVersion()
	if db.realm.db_version < DB_VERSION then
		db.realm.db_version = DB_VERSION
		db.realm.entries = {}
		self:Print("Database change - all entries were removed.")
	end
end

function GearGauge:OnEnable()
	self:CheckDatabaseVersion()
	UpdateData()
	self:RegisterEvent("PLAYER_ENTERING_WORLD")
	self:RegisterEvent("PLAYER_LEVEL_UP")
	self:RegisterEvent("PLAYER_TARGET_CHANGED")
	self:RegisterEvent("UNIT_INVENTORY_CHANGED")
end

function GearGauge:OnDisable()
	self:UnregisterEvent("PLAYER_ENTERING_WORLD")
	self:UnregisterEvent("PLAYER_LEVEL_UP")
	self:UnregisterEvent("PLAYER_TARGET_CHANGED")
	self:UnregisterEvent("UNIT_INVENTORY_CHANGED")
end

-------------------------------------------------------------------------------
-- Event handlers
-------------------------------------------------------------------------------
function GearGauge:PLAYER_ENTERING_WORLD()
	self:CheckDatabaseVersion()
	self:UnregisterEvent("PLAYER_ENTERING_WORLD")
end

function GearGauge:PLAYER_LEVEL_UP()
	UpdateData()
end

function GearGauge:PLAYER_TARGET_CHANGED()
	self:Scan("target")
end

do
	local last_update = 0
	local updater = CreateFrame("Frame", nil, UIParent)

	updater:Hide()
	updater:SetScript("OnUpdate",
			  function(self, elapsed)
				  last_update = last_update + elapsed
				  if last_update > 0.25 then
					  UpdateData()
					  if tooltip then
						  ShowTooltip(tipdata.anchor, tipdata.unit, tipdata.center, true)
					  end
					  self:Hide()
				  end
			  end)

	function GearGauge:UNIT_INVENTORY_CHANGED(event, unit)
		if unit ~= "player" then return end
		if not updater:IsVisible() then
			last_update = 0
			updater:Show()
		end
	end
end	-- do

-------------------------------------------------------------------------------
-- Configuration
-------------------------------------------------------------------------------
local GetOptions, GetScanOptions
do
	local TOOLTIP_COLORING_VALUES = {
		["Line"]	= L["Line"],
		["Text"]	= L["Text"],
		["Both"]	= L["Line and Text"],
	}
	local options, scan_options

	function GetOptions()
		if options then return options end

		options = {
			type	= "group",
			name	= MODNAME,
			args = {
				about = {
					order	= 1,
					type	= "description",
					name	= GAME_VERSION_LABEL.." "..version,
				},
				spacer_1 = {
					order		= 2,
					type		= "description",
					name		= " ",
					cmdHidden	= true
				},
				minimap_icon = {
					order	= 2,
					type	= "toggle",
					name	= L["MINIMAP_ICON_NAME"],
					desc	= L["MINIMAP_ICON_DESC"],
					get	= function() return not db.global.minimap_icon.hide end,
					set	= function(info, val)
							  db.global.minimap_icon.hide = not val

							  if val then
								  LibStub("LibDBIcon-1.0"):Show(MODNAME)
							  else
								  LibStub("LibDBIcon-1.0"):Hide(MODNAME)
							  end
						  end,
				},
				-------------------------------------------------------------------------------
				-- Documentation text
				-------------------------------------------------------------------------------
				documentation = {
					order	= 3,
					type	= "group",
					name	= L["ADDON_DOCUMENTATION_NAME"],
					desc	= L["ADDON_DOCUMENTATION_DESC"],
					args = {
						header1 = {
							order	= 1,
							type	= "header",
							name	= L["ADDON_DOCUMENTATION_NAME"],
						},
						desc1 = {
							order	= 2,
							type	= "header",
							name	= L["ADDON_DOCUMENTATION_DESC"],
						}
					},
				},
				-------------------------------------------------------------------------------
				-- Scanning options.
				-------------------------------------------------------------------------------
				scanning = {
					order	= 4,
					type	= "group",
					name	= L["Scanning"],
					desc	= L["SCANNING_OPTIONS_DESC"],
					args = {
						scanning_desc = {
							order	= 1,
							type	= "description",
							name	= L["SCANNING_OPTIONS_DESC"].."\n",
						},
						minlev = {
							order	= 2,
							type	= "range",
							name	= L["Minimum Level"],
							desc	= L["SCAN_MINLEVEL_DESC"],
							min	= 1,
							max	= MAX_PLAYER_LEVEL,
							step	= 1,
							get	= function() return db.global.scanning.minlevel end,
							set	= function(info, val)
									  db.global.scanning.minlevel = math.max(1, math.min(db.global.scanning.maxlevel, val))
								  end,
						},
						maxlev = {
							order	= 3,
							type	= "range",
							name	= L["Maximum Level"],
							desc	= L["SCAN_MAXLEVEL_DESC"],
							min	= 1,
							max	= MAX_PLAYER_LEVEL,
							step	= 1,
							get	= function() return db.global.scanning.maxlevel end,
							set	= function(info, val)
									  db.global.scanning.maxlevel = math.min(MAX_PLAYER_LEVEL, math.max(db.global.scanning.minlevel, val))
								  end,
						},
						delay = {
							order	= 4,
							type	= "range",
							name	= L["Delay"],
							desc	= L["SCAN_DELAY_DESC"],
							min	= 1,
							max	= 60,
							step	= 1,
							get	= function() return db.global.scanning.delay end,
							set	= function(info, val) db.global.scanning.delay = math.min(60, math.max(1, val)) end,
						},
						persistence = {
							order	= 5,
							type	= "range",
							name	= L["Persistence"],
							desc	= L["SCAN_PERSISTENCE_DESC"],
							min	= 1,
							max	= 30,
							step	= 1,
							get	= function() return db.global.scanning.persistence end,
							set	= function(info, val) db.global.scanning.persistence = math.min(30, math.max(1, val)) end,
						},
						spacer_1 = {
							order		= 6,
							type		= "description",
							name		= " ",
							cmdHidden	= true
						},
						notify = {
							order	= 7,
							type	= "toggle",
							name	= L["Notify"],
							desc	= L["SCAN_NOTIFY_DESC"],
							get	= function() return db.global.scanning.notify end,
							set	= function() db.global.scanning.notify = not db.global.scanning.notify end,
						},
						enemies = {
							order	= 8,
							type	= "toggle",
							name	= L["Scan Enemies"],
							desc	= L["SCAN_ENEMIES_DESC"],
							get	= function() return db.global.scanning.enemies end,
							set	= function() db.global.scanning.enemies = not db.global.scanning.enemies end,
						},
					},
				},
				-------------------------------------------------------------------------------
				-- Tooltip options.
				-------------------------------------------------------------------------------
				tooltip = {
					order	= 5,
					type	= "group",
					name	= L["Tooltip"],
					desc	= L["TOOLTIP_OPTIONS_DESC"],
					args = {
						tooltip_desc = {
							order	= 1,
							type	= "description",
							name	= L["TOOLTIP_OPTIONS_DESC"].."\n",
						},
--						scale = {
--							order	= 1,
--							type	= "range",
--							name	= L["Scale"],
--							desc	= L["TOOLTIP_SCALE_DESC"],
--							min	= 0.5,
--							max	= 2.0,
--							step	= 0.05,
--							get	= function() return db.global.tooltip.scale end,
--							set	= function(info, val) db.global.tooltip.scale = math.min(2.0, math.max(0.5, val)) end,
--						},
						timer = {
							order	= 2,
							type	= "range",
							name	= L["Timer"],
							desc	= L["TOOLTIP_TIMER_DESC"],
							min	= 0.25,
							max	= 3.0,
							step	= 0.05,
							get	= function() return db.global.tooltip.timer end,
							set	= function(info, val) db.global.tooltip.timer = math.min(3.0, math.max(0.25, val)) end,
						},
						coloring = {
							order	= 3,
							type	= "select",
							name	= L["Coloring"],
							desc	= L["TOOLTIP_COLORING_DESC"],
							get	= function() return db.global.tooltip.coloring end,
							set	= function(info, val) db.global.tooltip.coloring = val end,
							values	= TOOLTIP_COLORING_VALUES,
						},
					},
				},
			},
		}
		return options
	end

	function GearGauge:SetupOptions()
		LibStub("AceConfigRegistry-3.0"):RegisterOptionsTable(MODNAME, GetOptions())
		self.optionsFrame = LibStub("AceConfigDialog-3.0"):AddToBlizOptions(MODNAME)
	end
end	-- do


local function TimeStr(seconds)
	local tm = seconds
	local timeval = tostring(math.floor(tm % 60)).."s"
	local val

	if (tm >= 60) then
		tm = tm / 60
		val = math.floor(tm % 60)

		if val > 0 then	timeval = tostring(val).."m "..timeval end

		if (tm >= 60) then
			tm = tm / 60
			val = math.floor(tm % 24)

			if val > 0 then	timeval = tostring(val).."h "..timeval end

			if (tm >= 24) then
				tm = tm / 24
				timeval = tostring(math.floor(tm)).."d "..timeval
			end
		end
	end
	return timeval
end

-------------------------------------------------------------------------------
-- Subcommands
-------------------------------------------------------------------------------
local function ShowGear(arg1)
	-- Default to target if no argument was specified.
	if not arg1 or arg1 == "" then
		if not UnitExists("target") then
			GearGauge:Print(ERR_GENERIC_NO_TARGET)
			return
		end
		if not CanScan("target", false) then
			GearGauge:Print(ERR_CLIENT_LOCKED_OUT)
			return
		end
		ShowTooltip(UIParent, "target", true)
		return
	end
	local entry = FindEntry(arg1)

	if not entry then
		GearGauge:Print(string.format(ERR_BAD_PLAYER_NAME_S, arg1))
		return
	end
	ShowTooltip(UIParent, entry, true)
end

local function RemoveInfo(entry)
	local found_name = FindEntry(entry)

	if not found_name then
		GearGauge:Print(string.format(ERR_BAD_PLAYER_NAME_S, entry))
	else
		RemoveEntry(found_name, true)
	end
end

local ListEntries
do
	local list_entries = {}
	local name_list = {}

	local SORT_FUNCS = {
		["ClassSortDown"] = function(one, two) return ORDERED_CLASSES[one.class] < ORDERED_CLASSES[two.class] end,
		["ClassSortUp"] = function(one, two) return ORDERED_CLASSES[one.class] > ORDERED_CLASSES[two.class] end,
		["FactionSortDown"] = function(one, two) return one.faction < two.faction end,
		["FactionSortUp"] = function(one, two) return one.faction > two.faction end,
		["GuildSortDown"] = function(one, two)
					    if not one.guild and not two.guild then return false end
					    if not one.guild then return true end
					    if not two.guild then return false end
					    return one.guild < two.guild
				    end,
		["GuildSortUp"] = function(one, two)
					  if not one.guild and not two.guild then return false end
					  if not two.guild then return true end
					  if not one.guild then return false end
					  return one.guild > two.guild
				  end,
		["LevelSortDown"] = function(one, two) return one.level < two.level end,
		["LevelSortUp"] = function(one, two) return one.level > two.level end,
		["NameSortDown"] = function(one, two) return name_list[one] < name_list[two] end,
		["NameSortUp"] = function(one, two) return name_list[one] > name_list[two] end,
		["RaceSortDown"] = function(one, two) return one.race < two.race end,
		["RaceSortUp"] = function(one, two) return one.race > two.race end,
		["ScanSortDown"] = function(one, two) return one.scanned > two.scanned end,
		["ScanSortUp"] = function(one, two) return one.scanned < two.scanned end,
	}
	local CLASS_COLORS = {}
	local COLOR_TABLE = CUSTOM_CLASS_COLORS or RAID_CLASS_COLORS
	local player_color = COLOR_TABLE[select(2, UnitClass("player"))]

	for k, v in pairs(COLOR_TABLE) do
		CLASS_COLORS[k] = string.format("%2x%2x%2x", v.r * 255, v.g * 255, v.b * 255)
	end
	local game_font_color = string.format("|cff%2x%2x%2x", NORMAL_FONT_COLOR.r * 255, NORMAL_FONT_COLOR.g * 255, NORMAL_FONT_COLOR.b * 255)

	local list, right_pane
	local prev_name, prev_sort

	local function ListPanelIsOpen()
		return list and right_pane
	end

	local function CloseListPanel()
		list = QTC:Release(list)
		footer = QTC:Release(right_pane)
		tooltip = QTC:Release(tooltip)
	end

	function GearGauge:RefreshList()
		if ListPanelIsOpen() then
			CloseListPanel()
			ListEntries(prev_sort)
		end
	end

	local function RightPane_Click(event, cell, arg, button)
		if not arg then
			QTC.OnLeave(event, cell, arg)
			CloseListPanel()
			return
		end
		local dir = (button == "LeftButton") and "Down" or "Up"

		prev_sort = arg..dir
		CloseListPanel()
		ListEntries(prev_sort)
	end

	local function ListClick(event, cell, arg, button)
		if prev_name == arg then
			prev_name = nil
			tooltip = QTC:Release(tooltip)
			return
		end

		if IsAltKeyDown() and button == "RightButton" then
			CloseListPanel()
			RemoveEntry(arg, true)
			ListEntries(prev_sort)
			return
		end
		prev_name = arg
		ShowGear(arg)		-- FIXME: This shouldn't have to be called twice to get correct gem/enchant information.
		ShowGear(arg)
	end

	function ListEntries(func)
		if ListPanelIsOpen() then
			CloseListPanel()
			return
		end
		wipe(list_entries)
		wipe(name_list)

		for entry in pairs(db.realm.entries) do
			name_list[db.realm.entries[entry]] = entry
			tinsert(list_entries, db.realm.entries[entry])
		end

		if #list_entries < 1 then
			GearGauge:Print(EMPTY)
			return
		end
		local now = time()
		local player_faction = UnitFactionGroup("player")

		if not func or func == "" or not SORT_FUNCS[func] then func = "NameSortDown" end

		tsort(list_entries, SORT_FUNCS[func])

		list = QTC:Acquire(MODNAME.."ListFrame", 7)
		list:SetFrameStrata("FULLSCREEN_DIALOG")
		list:SetCallback("OnMouseUp", ListClick)
		list:ClearAllPoints()
		list:SetPoint("CENTER", UIParent, 0, 0)
		list:Clear()

		local line, column

		list:AddNormalHeader(NAME, LEVEL, RACE, CLASS, FACTION, GUILD, L["TIMESTAMP"])
		list:AddSeparator(1, COLOR_TABLE[UnitClass("player")])

		for idx, field in ipairs(list_entries) do
			local entry = list_entries[idx]
			local enemy = entry.faction ~= player_faction
			local name = name_list[entry]

			line, column = list:AddNormalLine()
			line, column = list:SetCell(line, 1, name, name)
			line, column = list:SetNormalCell(line, column, entry.level)
			line, column = list:SetNormalCell(line, column, entry.race)

			local class_english = ORDERED_CLASSES[entry.class]
			local class_name = db.global.localized_classes[class_english] or class_english

			if not class_name or not CLASS_COLORS[class_english] then
				line, column = list:SetNormalCell(line, column, entry.class)
			else
				line, column = list:SetNormalCell(line, column, "|cff"..CLASS_COLORS[class_english]..class_name)
			end
			line, column = list:SetNormalCell(line, column, (enemy and COLOR_RED or "")..entry.faction)
			line, column = list:SetNormalCell(line, column, entry.guild or "")
			line, column = list:SetNormalCell(line, column, TimeStr(now - entry.scanned))
		end
		list:UpdateScrolling(300)

		right_pane = QTC:Acquire(MODNAME.."ListFrameRight", 1, "CENTER")
		right_pane:SetFrameStrata("FULLSCREEN_DIALOG")
		right_pane:SetCallback("OnMouseUp", RightPane_Click)
		right_pane:ClearAllPoints()
		right_pane:SetPoint("TOPLEFT", list, "TOPRIGHT", 0, 0)
		right_pane:SetPoint("BOTTOMLEFT", list, "BOTTOMRIGHT", 0, 0)
		right_pane:Clear()

		local line, column

		right_pane:AddNormalHeader(L["SORT_ENTRIES_STR"])
		right_pane:AddSeparator(1, unpack(player_color))

		line, column = right_pane:AddLine()
		right_pane:SetCell(line, column, game_font_color..NAME, "NameSort")

		line, column = right_pane:AddLine()
		right_pane:SetCell(line, column, game_font_color..LEVEL, "LevelSort")

		line, column = right_pane:AddLine()
		right_pane:SetCell(line, column, game_font_color..RACE, "RaceSort")

		line, column = right_pane:AddLine()
		right_pane:SetCell(line, column, game_font_color..CLASS, "ClassSort")

		line, column = right_pane:AddLine()
		right_pane:SetCell(line, column, game_font_color..FACTION, "FactionSort")

		line, column = right_pane:AddLine()
		right_pane:SetCell(line, column, game_font_color..GUILD, "GuildSort")

		line, column = right_pane:AddLine()
		right_pane:SetCell(line, column, game_font_color..L["TIMESTAMP"], "ScanSort")

		right_pane:AddSeparator(1, unpack(player_color))

		for i = 1, 7 do
			right_pane:AddNormalLine(" ")
		end
		line, column = right_pane:AddNormalLine()
		right_pane:SetCell(line, column, string.format(L["PLAYER_COUNT_STR"], #list_entries), "CENTER")

		right_pane:AddSeparator(1, unpack(player_color))
		right_pane:AddSeparator(1, unpack(player_color))

		line, column = right_pane:AddLine()
		right_pane:SetCell(line, 1, game_font_color..CLOSE, "CENTER")

		right_pane:AddSeparator(1, unpack(player_color))
		right_pane:AddSeparator(1, unpack(player_color))

		list:Show()
		right_pane:Show()

		if TipTac and TipTac.AddModifiedTip then
			TipTac:AddModifiedTip(list, true)
			TipTac:AddModifiedTip(right_pane, true)
		end
	end
end	-- do

local function PurgeEntries()
	local count = CheckExpiration()

	GearGauge:Print(string.format(L["PURGE_MSG"], count))
end

local function RateEntry(entry)
	local name = FindEntry(entry)

	if not name then
		GearGauge:Print(string.format(ERR_BAD_PLAYER_NAME_S, entry))
		return
	end
	local info = db.realm.entries[name]
	local items = AcquireTable("RateEntry")
	local rating = BuildGearData(info, items)
	local equip_level = AverageItemLevels(items)

	local enemies = info.faction ~= UnitFactionGroup("player")
	local color_str = ""
	if enemies then color_str = COLOR_RED end

	GearGauge:Print(color_str..name.." "..RATING..": "..string.format("%.2f", rating)..". "..ITEMS.." "..LEVEL..": "..equip_level)
	ReleaseItems(items)
end

--[===[@debug@
local ShowTable
do
	local table_lookup = {}

	function ShowTables(arg)
		if used_tables then
			GearGauge:Print(string.format("Used tables: %d", used_tables))
			wipe(table_lookup)

			for k, v in pairs(table_locs) do
				if table_lookup[v] then
					table_lookup[v] = table_lookup[v] + 1
				else
					table_lookup[v] = 1
				end
			end

			for k, v in pairs(table_lookup) do
				debug(string.format("%s: %d", k, v))
			end
		end

		if cached_tables then
			GearGauge:Print(string.format("Cached tables: %d", cached_tables))
		end
	end
end	-- do
--@end-debug@]===]

local function TogglePanel()
	if tooltip then
		tooltip = QTC:Release(tooltip)
	else
		ShowTooltip(UIParent, "player", true)
	end
end

local function HidePanel()
	if tooltip then
		tooltip = QTC:Release(tooltip)
	end
end

function OpenConfig()
	local options_frame = InterfaceOptionsFrame

	if options_frame:IsVisible() then
		options_frame:Hide()
	else
		InterfaceOptionsFrame_OpenToCategory(GearGauge.optionsFrame)
	end
end

-------------------------------------------------------------------------------
-- Command table and parser.
-------------------------------------------------------------------------------
do
	local COMMANDS = {
		["config"]	= OpenConfig,
		["hide"]	= HidePanel,
		["list"]	= ListEntries,
		["panel"]	= TogglePanel,
		["purge"]	= PurgeEntries,
		["rate"]	= RateEntry,
		["remove"]	= RemoveInfo,
		["show"]	= ShowGear,
		--[===[@debug@
		["tables"]	= ShowTables,
		--@end-debug@]===]
	}

	function GearGauge:ShowCommands()
		local str = ""
		for k, v in pairs(COMMANDS) do
			str = string.format("%s\n  %s", str, k)
		end
		self:Print(str)
	end
	-------------------------------------------------------------------------------
	-- Separates input into two parts, by assigning the first word to arg1 and
	-- the rest of the string to arg2
	-------------------------------------------------------------------------------
	local function StrSplit(input)
		if not input then return nil, nil end
		local arg1, arg2, var1

		arg1, var1 = input:match("^([^%s]+)%s*(.*)$")
		arg1 = (arg1 and arg1:lower() or input:lower())

		if var1 then
			local var2
			arg2, var2 = var1:match("^([^%s]+)%s*(.*)$")
			arg2 = (arg2 and arg2:lower() or var1:lower())
		end
		return arg1, arg2
	end

	function GearGauge:SlashCommand(input)
		local cmd, arg1, arg2, arg3
		local str = input

		cmd, str = StrSplit(str)
		arg1, str = StrSplit(str)
		arg2, arg3 = StrSplit(str)

		if not cmd or (cmd and cmd:trim() == "") then
			self:ShowCommands()
			return
		end

		if not COMMANDS[cmd] then
			self:Print(string.format(L["INVALID_CMD_STR"], cmd))
			self:ShowCommands()
		else
			COMMANDS[cmd](arg1, arg2, arg3)
		end
	end
end	-- do
